@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import argparse
 import functools
 import operator
@@ -7,7 +9,7 @@ import re
 import sys
 import warnings
 from functools import singledispatch
-from typing import Union
+from typing import Iterator
 
 import h5py
 import numpy as np
@@ -71,8 +73,8 @@ def dump(file: h5py.File, data: dict, root: str = "/"):
         file[join(root, path)] = _dict_get(data, path)
 
 
-def abspath(path):
-    r"""
+def abspath(path: str) -> str:
+    """
     Return absolute path.
 
     :param str path: A HDF5-path.
@@ -81,11 +83,12 @@ def abspath(path):
     return posixpath.normpath(posixpath.join("/", path))
 
 
-def join(*args, root=False):
-    r"""
+def join(*args, root: bool = False) -> str:
+    """
     Join path components.
 
     :param list args: Piece of a path.
+    :param root: Prepend the output with the root ``"/"``.
     :return: The concatenated path.
     """
 
@@ -103,13 +106,16 @@ def join(*args, root=False):
     return posixpath.join(*lst)
 
 
-def getdatapaths(file, root: str = "/"):
+def getdatapaths(file, root: str = "/") -> list[str]:
     """
-    Get paths to all datasets and groups that contain attributes.
+    Get paths to all:
+
+    -   Datasets, and
+    -   Groups that contain attributes.
 
     :param file: A HDF5-archive.
     :param root: Start at a certain point along the path-tree.
-    :return: ``list[str]``.
+    :return: List of paths.
     """
     return list(getdatasets(file, root=root)) + list(getgroups(file, root=root, has_attrs=True))
 
@@ -124,7 +130,7 @@ def getgroups(
     :param root: Start at a certain point along the path-tree.
     :param has_attrs: Return only groups that have attributes.
     :param int max_depth: Set a maximum depth beyond which groups are folded.
-    :return: ``list[str]``.
+    :return: List of paths.
     """
 
     keys = []
@@ -145,7 +151,9 @@ def getgroups(
     return keys
 
 
-def getdatasets(file, root="/", max_depth=None, fold=None):
+def getdatasets(
+    file: h5py.File, root: str = "/", max_depth: int = None, fold: str | list[str] = None
+) -> Iterator:
     r"""
     Iterator to transverse all datasets in a HDF5-archive.
     One can choose to fold (not transverse deeper than):
@@ -153,11 +161,11 @@ def getdatasets(file, root="/", max_depth=None, fold=None):
     -   Groups deeper than a certain ``max_depth``.
     -   A (list of) specific group(s).
 
-    :param h5py.File file: A HDF5-archive.
-    :param str root: Start a certain point along the path-tree.
-    :param int max_depth: Set a maximum depth beyond which groups are folded.
-    :param list fold: Specify groups that are folded.
-    :return: Iterator.
+    :param file: A HDF5-archive.
+    :param root: Start a certain point along the path-tree.
+    :param max_depth: Set a maximum depth beyond which groups are folded.
+    :param fold: Specify groups that are folded.
+    :return: Iterator to paths.
 
     :example:
 
@@ -205,58 +213,6 @@ def getdatasets(file, root="/", max_depth=None, fold=None):
     return _getpaths(file, root)
 
 
-def getpaths(data, root="/", max_depth=None, fold=None):
-    r"""
-    Iterator to transverse all datasets in HDF5-archive.
-    One can choose to fold (not transverse deeper than):
-
-    -   Groups deeper than a certain ``max_depth``.
-    -   A (list of) specific group(s).
-
-    :param h5py.File data: A HDF5-archive.
-    :param str root: Start a certain point along the path-tree.
-    :param int max_depth: Set a maximum depth beyond which groups are folded.
-    :param list fold: Specify groups that are folded.
-    :return: Iterator.
-
-    :example:
-
-        Consider this file:
-
-        .. code-block:: bash
-
-            /path/to/first/a
-            /path/to/first/b
-            /data/c
-            /data/d
-            /e
-
-        Calling:
-
-        .. code-block:: python
-
-            with h5py.File('...', 'r') as data:
-
-                for path in GooseHDF5.getpaths(data, max_depth=2, fold='/data'):
-                    print(path)
-
-        Will print:
-
-        .. code-block:: bash
-
-            /path/to/...
-            /data/...
-            /e
-
-        The ``...`` indicate that it concerns a folded group, not a dataset.
-        Here, the first group was folded because of the maximum depth, and the second because it was
-        specifically requested to be folded.
-    """
-
-    warnings.warn("getpaths() is deprecated, use getdatasets().", warnings.DeprecationWarning)
-    return getdatasets(data, root, max_depth, fold)
-
-
 def _getpaths(file, root):
     r"""
     Specialization for :py:func:`getpaths`.
@@ -518,7 +474,7 @@ def _create_groups(file, paths):
             file.create_group(group)
 
 
-def _copy(source, dest, source_paths, dest_paths):
+def _copy(source, dest, source_paths, dest_paths, expand_soft):
     """
     Copy paths recursively.
     """
@@ -533,11 +489,18 @@ def _copy(source, dest, source_paths, dest_paths):
         # (main function checks existence before, so this can be the only reason)
         if dest_path in dest:
             continue
+
+        if not expand_soft:
+            link = source.get(source_path, getlink=True)
+            if isinstance(link, h5py.SoftLink):
+                dest[dest_path] = h5py.SoftLink(link.path)
+                continue
+
         group = posixpath.split(dest_path)[0]
         source.copy(source_path, dest[group], posixpath.split(dest_path)[1])
 
 
-def _copy_attrs(source, dest, source_paths, dest_paths):
+def _copy_attrs(source, dest, source_paths, dest_paths, expand_soft):
 
     if len(source_paths) == 0:
         return 0
@@ -546,6 +509,12 @@ def _copy_attrs(source, dest, source_paths, dest_paths):
 
     for source_path, dest_path in zip(source_paths, dest_paths):
 
+        if not expand_soft:
+            link = source.get(source_path, getlink=True)
+            if isinstance(link, h5py.SoftLink):
+                dest[dest_path] = h5py.SoftLink(link.path)
+                continue
+
         source_group = source[source_path]
 
         if dest_path not in dest:
@@ -564,6 +533,8 @@ def copy(
     dest_datasets: list[str] = None,
     root: str = None,
     recursive: bool = True,
+    skip: bool = False,
+    expand_soft: bool = True,
 ):
     """
     Copy groups/datasets from one HDF5-archive ``source`` to another HDF5-archive ``dest``.
@@ -577,6 +548,8 @@ def copy(
     :param dest_datasets: List of dataset-paths in ``dest``, defaults to ``source_datasets``.
     :param root: Path prefix for all ``dest_datasets``.
     :param recursive: If the source is a group, copy all objects within that group recursively.
+    :param skip: Skip datasets that are not present in source.
+    :param expand_soft: Copy the underlying data of a link, or copy as link with the same path.
     """
 
     if len(source_datasets) == 0:
@@ -589,6 +562,11 @@ def copy(
 
     dest_datasets = np.array(dest_datasets)
 
+    if skip:
+        keep = [i in source for i in source_datasets]
+        source_datasets = source_datasets[keep]
+        dest_datasets = dest_datasets[keep]
+
     if root:
         dest_datasets = np.array([join(root, path, root=True) for path in dest_datasets])
 
@@ -599,17 +577,22 @@ def copy(
     isgroup = np.array([isinstance(source[path], h5py.Group) for path in source_datasets])
 
     if recursive:
-        _copy(source, dest, source_datasets[isgroup], dest_datasets[isgroup])
+        _copy(
+            source, dest, source_datasets[isgroup], dest_datasets[isgroup], expand_soft=expand_soft
+        )
 
     _copy(
         source,
         dest,
         source_datasets[np.logical_not(isgroup)],
         dest_datasets[np.logical_not(isgroup)],
+        expand_soft=expand_soft,
     )
 
     if not recursive:
-        _copy_attrs(source, dest, source_datasets[isgroup], dest_datasets[isgroup])
+        _copy_attrs(
+            source, dest, source_datasets[isgroup], dest_datasets[isgroup], expand_soft=expand_soft
+        )
 
 
 def copydatasets(
@@ -635,7 +618,7 @@ def copydatasets(
     :param root: Path prefix for all ``dest_datasets``.
     """
 
-    warnings.warn("copydatasets() is deprecated, use copy().", warnings.DeprecationWarning)
+    warnings.warn("copydatasets() is deprecated, use copy().", category=DeprecationWarning)
 
     return copy(source, dest, source_datasets, dest_datasets, root)
 
@@ -699,7 +682,7 @@ def _equal_value(a, b):
         else:
             return False
 
-    return list(a[...]) == list(b[...])
+    return np.all(np.array(list(a[...])) == np.array(list(b[...])))
 
 
 def _equal(a, b, attrs, matching_dtype):
@@ -834,11 +817,10 @@ def _compare_paths(
 
 @singledispatch
 def compare(
-    a: Union[str, h5py.File],
-    b: Union[str, h5py.File],
+    a: str | h5py.File,
+    b: str | h5py.File,
     paths_a: list[str] = None,
     paths_b: list[str] = None,
-    root: str = None,
     attrs: bool = True,
     matching_dtype: bool = False,
 ):
@@ -1016,20 +998,46 @@ def copy_dataset(source, dest, paths, compress=False, double_to_float=False):
             dest[path].attrs[key] = source[path].attrs[key]
 
 
-def print_plain(source, paths: list[str]):
+def _linktype2str(source: h5py.File | h5py.Group, path: str) -> str:
+
+    dset = source.get(path, getlink=True)
+
+    if isinstance(dset, h5py.SoftLink):
+        return "soft"
+
+    if isinstance(dset, h5py.HardLink):
+        return "hard"
+
+    if isinstance(dset, h5py.ExternalLink):
+        return "external"
+
+    return "??"
+
+
+def print_plain(source, paths: list[str], show_links: bool = False):
     r"""
     Print the paths to all datasets (one per line).
     :param paths: List of paths.
+    :param show_links: Show the path the link points to.
     """
 
+    if show_links:
+        for path in paths:
+            if isinstance(source.get(path, getlink=True), h5py.SoftLink):
+                print(path + " -> " + source.get(path, getlink=True).path)
+            else:
+                print(path)
+        return
+
     for path in paths:
         print(path)
 
 
-def print_info(source, paths: list[str]):
+def print_info(source, paths: list[str], link_type: bool = False):
     r"""
     Print the paths to all datasets (one per line), including type information.
     :param paths: List of paths.
+    :param link_type: Include the link-type in the output.
     """
 
     def has_attributes(lst):
@@ -1060,12 +1068,15 @@ def print_info(source, paths: list[str]):
             out["dtype"] += ["-"]
             out["attrs"] += ["-"]
 
+    if link_type:
+        out["link"] = [_linktype2str(source, path) for path in paths]
+
     width = {}
     for key in out:
         width[key] = max(len(i) for i in out[key])
         width[key] = max(width[key], len(key))
 
-    fmt = "{0:%ds} {1:%ds} {2:%ds} {3:%ds}" % (
+    fmt = "{path:%ds} {size:%ds} {shape:%ds} {dtype:%ds}" % (
         width["path"],
         width["size"],
         width["shape"],
@@ -1073,29 +1084,16 @@ def print_info(source, paths: list[str]):
     )
 
     if has_attributes(out["attrs"]):
-        fmt += " {4:%ds}" % width["attrs"]
+        fmt += " {attrs:%ds}" % width["attrs"]
 
-    print(fmt.format("path", "size", "shape", "dtype", "attrs"))
-    print(
-        fmt.format(
-            "=" * width["path"],
-            "=" * width["size"],
-            "=" * width["shape"],
-            "=" * width["dtype"],
-            "=" * width["attrs"],
-        )
-    )
+    if "link" in out:
+        fmt += " {link:%ds}" % width["link"]
+
+    print(fmt.format(**{key: key for key in out}))
+    print(fmt.format(**{key: "=" * width[key] for key in out}))
 
     for i in range(len(out["path"])):
-        print(
-            fmt.format(
-                out["path"][i],
-                out["size"][i],
-                out["shape"][i],
-                out["dtype"][i],
-                out["attrs"][i],
-            )
-        )
+        print(fmt.format(**{key: out[key][i] for key in out}))
 
 
 def print_attribute(source, paths: list[str]):
@@ -1233,14 +1231,21 @@ def _G5list_parser():
     Return parser for :py:func:`G5list`.
     """
 
-    desc = "List datasets (or groups of datasets) in a HDF5-file"
+    desc = """List datasets (or groups of datasets) in a HDF5-file.
+    Note that if you have a really big file the ``--layer`` option can be much faster than
+    searching the entire file.
+    """
     parser = argparse.ArgumentParser(description=desc)
+    parser.add_argument("--min-attrs", type=int, help="Filter based on min. number of attributes")
+    parser.add_argument("--link-type", action="store_true", help="Print the link-type")
     parser.add_argument("-D", "--datasets", action="store_true", help="Print only datasets")
     parser.add_argument("-d", "--max-depth", type=int, help="Maximum depth to display")
     parser.add_argument("-f", "--fold", action="append", help="Fold paths")
     parser.add_argument("-i", "--info", action="store_true", help="Print info: shape, dtype")
+    parser.add_argument("-L", "--layer", type=str, help="Print paths at a specific layer")
     parser.add_argument("-l", "--long", action="store_true", help="--info but with attributes")
     parser.add_argument("-r", "--root", default="/", help="Start somewhere in the path-tree")
+    parser.add_argument("-s", "--links", action="store_true", help="Show destination of soft links")
     parser.add_argument("-v", "--version", action="version", version=version)
     parser.add_argument("source")
     return parser
@@ -1260,17 +1265,32 @@ def G5list(args: list[str]):
 
     with h5py.File(args.source, "r") as source:
 
-        paths = list(getdatasets(source, root=args.root, max_depth=args.max_depth, fold=args.fold))
-        if not args.datasets:
-            paths += getgroups(source, root=args.root, has_attrs=True, max_depth=args.max_depth)
-        paths = sorted(list(set(paths)))
+        if args.layer is not None:
+            paths = sorted(join(args.layer, i, root=True) for i in source[args.layer])
+        else:
+            paths = list(
+                getdatasets(source, root=args.root, max_depth=args.max_depth, fold=args.fold)
+            )
+            if not args.datasets:
+                paths += getgroups(source, root=args.root, has_attrs=True, max_depth=args.max_depth)
+            paths = sorted(list(set(paths)))
+
+        if args.min_attrs is not None:
+            rm = []
+            for path in paths:
+                if path in source:
+                    source[path]
+                    if len(source[path].attrs) < args.min_attrs:
+                        rm.append(path)
+            for path in rm:
+                paths.remove(path)
 
         if args.info:
-            print_info(source, paths)
+            print_info(source, paths, link_type=args.link_type)
         elif args.long:
             print_attribute(source, paths)
         else:
-            print_plain(source, paths)
+            print_plain(source, paths, show_links=args.links)
 
 
 def _G5list_catch():
